/**\
|**|	CopyBlits.c
\**/

// compiler directives
#define USE_SMP 1

// includes
#include "CopyBlits.h"

// typdef's, defines, enums, etc.
#if USE_SMP
#	define GETPIXBASEADDR(h) ((*h)->baseAddr)
#	define DO_LOCKING 0	// Must be zero for MP
#else
#	define GETPIXBASEADDR(h) GetPixBaseAddr(h)
#	define DO_LOCKING 0
#endif

// local (static) variables
static CTabHandle gCTabHandle = nil;
static UInt16 gCTab16[256];
static UInt32 gCTab32[256];

// implementation

// Save color translation table
void Set_ColorTable(CTabHandle pCTabHandle)
{
	if (gCTabHandle != pCTabHandle)
	{
		gCTabHandle = pCTabHandle;
		if (pCTabHandle)
		{
			if ((*pCTabHandle)->ctSize >= 255)
			{
				short index;
				for (index = 0;index < 256;index++)
				{
					RGBColor tRGBColor = (**pCTabHandle).ctTable[index].rgb;
					gCTab32[index] = 	((tRGBColor.red >> 8) << 16) +
										((tRGBColor.green >> 8) << 8) +
										((tRGBColor.blue >> 8) << 0);
					gCTab16[index] = 	((tRGBColor.red >> 11) << 10) +
										((tRGBColor.green >> 11) << 5) +
										((tRGBColor.blue >> 11) << 0);
				}
			}
		}
	}
}

// Generic routine that uses pixel size to determ which optimized routine to use
// If you KNOW your depth you should call the approate routine yourself and skip
// the subroutine/switch overhead.

void CopyBlits(	const PixMapHandle srcPixMapHdl,
				const PixMapHandle dstPixMapHdl,
				const Rect *srcRect,
				const Rect *dstRect
)
{
	if ((*srcPixMapHdl)->pixelSize == 8)
	{
		switch ((*dstPixMapHdl)->pixelSize)
		{
			case	8:
//				CopyBlits8(srcPixMapHdl,dstPixMapHdl,srcRect,dstRect);	// old
				CopyBlits8_8(srcPixMapHdl,dstPixMapHdl,srcRect,dstRect);
				break;
			case	16:
				CopyBlits8_16(srcPixMapHdl,dstPixMapHdl,srcRect,dstRect);
				break;
			case	32:
				CopyBlits8_32(srcPixMapHdl,dstPixMapHdl,srcRect,dstRect);
				break;
			default:
#if !USE_SMP
				CopyBits((BitMap*) *srcPixMapHdl,(BitMap*) *dstPixMapHdl,srcRect,dstRect,0,nil);
#endif USE_SMP
				break;
		}
	}
#if !USE_SMP
	else
		CopyBits((BitMap*) *srcPixMapHdl,(BitMap*) *dstPixMapHdl,srcRect,dstRect,0,nil);
#endif USE_SMP
}

// This is one of the first non-asm blitters I ever saw. Not as fast as CopyBlits8_8 but I left
// it in just to show you the skanky nested switch/while code (see Duff's device). ewueee!
// (194 fps vs. 300 fps CopyBits)
void CopyBlits8	(	const PixMapHandle srcPixMapHdl,
					const PixMapHandle dstPixMapHdl,
					const Rect *srcRect,
					const Rect *dstRect
)
{
	UInt32			dstLeft,dstRight;
	UInt32			*srcRow,*dstRow;
	register UInt32	*srcPtr,*dstPtr;
	UInt32			leftMask,rightMask;
	UInt32			notLeftMask,notRightMask;

	UInt32			srcRowBytes;
	UInt32			dstRowBytes;
	UInt32			srcLeft;

	UInt32			dstLong;
	UInt32			dstLongs;
	UInt32			height;
	UInt32			offset;
	UInt32			timesCopy;

	// This assumes that the pixmap is already locked
	srcRow = (UInt32 *) (*srcPixMapHdl)->baseAddr;
	dstRow = (UInt32 *) (*dstPixMapHdl)->baseAddr;
	
	// get the bit offset to the src left edge
	srcLeft = (srcRect->left - (*srcPixMapHdl)->bounds.left) * (*srcPixMapHdl)->pixelSize;
		
	// offset the src ptr to the first long
	srcRow += srcLeft >> 5;
	
	// get the bit offset to the dst left and right edges
	dstLeft = (dstRect->left - (*dstPixMapHdl)->bounds.left) * (*dstPixMapHdl)->pixelSize;
	dstRight = (dstRect->right - (*dstPixMapHdl)->bounds.left) * (*dstPixMapHdl)->pixelSize;
	
	// get the number of middle longs to do minus the left edge long
	dstLongs = ((dstRight - dstLeft) >> 5) - 1;
	
	// offset the dst Ptr to the first long
	dstRow += dstLeft >> 5;

	// now compute left and right masks for the dst
	dstLeft &= 0x1f;
	leftMask = ( 1 << dstLeft ) - 1;
	notLeftMask = ~leftMask;
	
	dstRight &= 0x1f;
	notRightMask = ( 1 << dstRight ) - 1;
	rightMask = ~notRightMask;
	
	srcRowBytes = (*srcPixMapHdl)->rowBytes & 0x3fff;
	dstRowBytes = (*dstPixMapHdl)->rowBytes & 0x3fff;

	// offset the src and dst ptrs to the first row
	offset = (srcRect->top - (*srcPixMapHdl)->bounds.top) * srcRowBytes;
	srcRow = (UInt32*) ((Ptr) srcRow + offset);

	offset = (dstRect->top - (*dstPixMapHdl)->bounds.top) * dstRowBytes;
	dstRow = (UInt32*) ((Ptr) dstRow + offset);

	/* check if we need to do the left and right mask */
	if ( leftMask )
	{
		if ( notLeftMask == 0 )
		{
			leftMask = 0;
			dstLongs++;
		}
	}
		
	if ( rightMask )
	{
		if ( notRightMask == 0 )
		{
			rightMask = 0;
			dstLongs++;
		}
	}

	height = srcRect->bottom - srcRect->top;	// No scaling allowed

	//for ( ; height >= 0; --height )
	// changing the above 'for()' to the below 'while()' is what made this blit routine
	// faster than CopyBits - a speed improvement equal to all other changes I had made
	// previously.  (about 4 to 5 milliseconds, in case you were wondering)
	// Lesson: the true bottlenecks are not always the obvious ones
	while (height--)
	{
		srcPtr = srcRow;
		dstPtr = dstRow;
		
		/* do the masked left edge */
		if ( leftMask )
		{
			dstLong = *srcPtr++ & leftMask;
			dstLong |= *dstPtr & notLeftMask;
			*dstPtr++ = dstLong;
		}

		/* do the middle longs with Duff's device */
		timesCopy = (dstLongs + 15) >> 4;
		
		switch( dstLongs & 0xF )
		{
			case 0:	do
				{	*dstPtr++ = *srcPtr++;
			case 15:	*dstPtr++ = *srcPtr++;
			case 14:	*dstPtr++ = *srcPtr++;
			case 13:	*dstPtr++ = *srcPtr++;
			case 12:	*dstPtr++ = *srcPtr++;
			case 11:	*dstPtr++ = *srcPtr++;
			case 10:	*dstPtr++ = *srcPtr++;
			case 9:		*dstPtr++ = *srcPtr++;
			case 8:		*dstPtr++ = *srcPtr++;
			case 7:		*dstPtr++ = *srcPtr++;
			case 6:		*dstPtr++ = *srcPtr++;
			case 5:		*dstPtr++ = *srcPtr++;
			case 4:		*dstPtr++ = *srcPtr++;
			case 3:		*dstPtr++ = *srcPtr++;
			case 2:		*dstPtr++ = *srcPtr++;
			case 1:		*dstPtr++ = *srcPtr++;
				} while( --timesCopy > 0 );
		}

		/* do the masked right edge */
		if ( rightMask )
		{
			dstLong = *srcPtr & rightMask;
			dstLong |= *dstPtr & notRightMask;
			*dstPtr = dstLong;
		}
		
		/* bump to the next row */
		srcRow = (UInt32*) ((Ptr) srcRow + srcRowBytes);
		dstRow = (UInt32*) ((Ptr) dstRow + dstRowBytes);
	}
}

// my Fastest (non-asm) 8 to 8 bit blitter
// (354 fps vs. 300 fps CopyBits)
void CopyBlits8_8	(	const PixMapHandle srcPixMapHdl,
						const PixMapHandle dstPixMapHdl,
						const Rect *srcRect,
						const Rect *dstRect
)
{
	UInt32	doubleCount,count;
	UInt32	startFlags, endFlags;
	UInt8	*src, *dst;
	UInt32	srcSkip, dstSkip;
	UInt32	width,height;

	width = srcRect->right - srcRect->left;
	height = srcRect->bottom - srcRect->top;

	if (!width || !height)
		return;	// nothing to copy!

#if DO_LOCKING
	{
		GWorldFlags srcGWFlags = GetPixelsState(srcPixMapHdl);
		GWorldFlags dstGWFlags = GetPixelsState(dstPixMapHdl);

		if (0 == (srcGWFlags & pixelsLocked))	// if source not locked
			if (!LockPixels(srcPixMapHdl))		// and we can't lock it
				return;	// give up

		if (0 == (dstGWFlags & pixelsLocked))	// if dest not locked
			if (!LockPixels(dstPixMapHdl))		// and we can't lock it
				goto unlock;
#endif

	srcSkip = (*srcPixMapHdl)->rowBytes & 0x3FFF;
	dstSkip = (*dstPixMapHdl)->rowBytes & 0x3FFF;

	src = (UInt8*) GETPIXBASEADDR(srcPixMapHdl) +
		(srcSkip * (srcRect->top - (*srcPixMapHdl)->bounds.top)) +
		srcRect->left - (*srcPixMapHdl)->bounds.left;
	dst = (UInt8*) GETPIXBASEADDR(dstPixMapHdl) +
		(dstSkip * (dstRect->top - (*dstPixMapHdl)->bounds.top)) +
		dstRect->left - (*dstPixMapHdl)->bounds.left;

	startFlags = ((UInt8 *) (((UInt32) src + 7U) & ~7U)) - src;
	endFlags = ((UInt32) (src + width)) & 7U;
	doubleCount = ((width - startFlags) - endFlags) / 8;

	// pre-fix skip values
	srcSkip -= width;
	dstSkip -= width;

	while (height--)
	{
		if (startFlags & 1)	// byte align
			*dst++ = *src++;

		if (startFlags & 2)	// word align
		{
			*(UInt16*) dst = *(UInt16*) src;
			src += 2;
			dst += 2;
		}

		if (startFlags & 4)	// long align
		{
			*(UInt32*) dst = *(UInt32*) src;
			src += 4;
			dst += 4;
		}

		// copy doubles
		for (count = 0; count < doubleCount; count++)
		{
			*(double*) dst = *(double*) src;
			src += 8;
			dst += 8;
		}

		// copy remaining long
		if (endFlags & 4)
		{
			*(UInt32*) dst = *(UInt32*) src;
			src += 4;
			dst += 4;
		}
			
		// copy remaining word
		if (endFlags & 2)
		{
			*(UInt16*) dst = *(UInt16*) src;
			src += 2;
			dst += 2;
		}

		// copy remaining byte
		if (endFlags & 1)
			*dst++ = *src++;

		src += srcSkip;
		dst += dstSkip;
	}
#if DO_LOCKING
unlock:
		if (0 == (srcGWFlags & pixelsLocked))	// if source wasn't locked
			UnlockPixels(srcPixMapHdl);			// unlock it
		if (0 == (dstGWFlags & pixelsLocked))	// if dest wasn't locked
			UnlockPixels(dstPixMapHdl);			// unlock it
	}
#endif
}

// my Fastest (non-asm) 8 to 16 bit blitter
// (50 fps vs. 42 fps CopyBits)
void CopyBlits8_16	(	const PixMapHandle srcPixMapHdl,
						const PixMapHandle dstPixMapHdl,
						const Rect *srcRect,
						const Rect *dstRect
)
{
	UInt8	*src;
	UInt16	*dst;
	UInt32	count;
	UInt32	srcSkip, dstSkip;
	UInt32	width,height;

	Set_ColorTable((**srcPixMapHdl).pmTable);

	width = srcRect->right - srcRect->left;
	height = srcRect->bottom - srcRect->top;

	if (!width || !height)
		return;

#if DO_LOCKING
	{
		GWorldFlags srcGWFlags = GetPixelsState(srcPixMapHdl);
		GWorldFlags dstGWFlags = GetPixelsState(dstPixMapHdl);

		if (0 == (srcGWFlags & pixelsLocked))	// if source not locked
			if (!LockPixels(srcPixMapHdl))		// and we can't lock it
				return;	// give up

		if (0 == (dstGWFlags & pixelsLocked))	// if dest not locked
			if (!LockPixels(dstPixMapHdl))		// and we can't lock it
				goto unlock;
#endif

	srcSkip = (*srcPixMapHdl)->rowBytes & 0x3FFF;
	dstSkip = (*dstPixMapHdl)->rowBytes & 0x3FFF;

	src = (UInt8*) (GETPIXBASEADDR(srcPixMapHdl) +
		(srcSkip * (srcRect->top - (*srcPixMapHdl)->bounds.top)) +
		srcRect->left - (*srcPixMapHdl)->bounds.left);
	dst = (UInt16*) (GETPIXBASEADDR(dstPixMapHdl) +
		(dstSkip * (dstRect->top - (*dstPixMapHdl)->bounds.top)) +
		((dstRect->left - (*dstPixMapHdl)->bounds.left) << 1));

	// pre-fix skip values
	srcSkip -= width;
	dstSkip -= width << 1;

	while (height--)
	{
		for (count = 0; count < width; count++)
			*dst++ = gCTab16[*src++];
		src += srcSkip;
		dst = (UInt16*) (dstSkip + (Ptr) dst);
	}

#if DO_LOCKING
unlock:
		if (0 == (srcGWFlags & pixelsLocked))	// if source wasn't locked
			UnlockPixels(srcPixMapHdl);			// unlock it
		if (0 == (dstGWFlags & pixelsLocked))	// if dest wasn't locked
			UnlockPixels(dstPixMapHdl);			// unlock it
	}
#endif
}

// my Fastest (non-asm) 8 to 32 bit blitter
// (50 fps vs. 39 fps CopyBits)
void CopyBlits8_32	(	const PixMapHandle srcPixMapHdl,
						const PixMapHandle dstPixMapHdl,
						const Rect *srcRect,
						const Rect *dstRect
)
{
	register	UInt8	*src;
	register	UInt32	*dst;
	register	UInt32	count;
	register	UInt32	srcSkip, dstSkip;
	register	UInt32	width,height;

	Set_ColorTable((**srcPixMapHdl).pmTable);

	width = srcRect->right - srcRect->left;
	height = srcRect->bottom - srcRect->top;

	if (!width || !height)
		return;

#if DO_LOCKING
	{
		GWorldFlags srcGWFlags = GetPixelsState(srcPixMapHdl);
		GWorldFlags dstGWFlags = GetPixelsState(dstPixMapHdl);

		if (0 == (srcGWFlags & pixelsLocked))	// if source not locked
			if (!LockPixels(srcPixMapHdl))		// and we can't lock it
				return;	// give up

		if (0 == (dstGWFlags & pixelsLocked))	// if dest not locked
			if (!LockPixels(dstPixMapHdl))		// and we can't lock it
				goto unlock;
#endif

	srcSkip = (*srcPixMapHdl)->rowBytes & 0x3FFF;
	dstSkip = (*dstPixMapHdl)->rowBytes & 0x3FFF;

	src = (UInt8*) (GETPIXBASEADDR(srcPixMapHdl) +
		(srcSkip * (srcRect->top - (*srcPixMapHdl)->bounds.top)) +
		srcRect->left - (*srcPixMapHdl)->bounds.left);
	dst = (UInt32*) (GETPIXBASEADDR(dstPixMapHdl) +
		(dstSkip * (dstRect->top - (*dstPixMapHdl)->bounds.top)) +
		((dstRect->left - (*dstPixMapHdl)->bounds.left) << 2));

	// pre-fix skip values
	srcSkip -= width;
	dstSkip -= width << 2;

	while (height--)
	{
		for (count = 0; count < width; count++)
			*dst++ = gCTab32[*src++];
		src += srcSkip;
		dst = (UInt32*) (dstSkip + (Ptr) dst);
	}

#if DO_LOCKING
unlock:
		if (0 == (srcGWFlags & pixelsLocked))	// if source wasn't locked
			UnlockPixels(srcPixMapHdl);			// unlock it
		if (0 == (dstGWFlags & pixelsLocked))	// if dest wasn't locked
			UnlockPixels(dstPixMapHdl);			// unlock it
	}
#endif
}
